# Note: this file was auto-generated by make2build 0.18

use Module::Build;

my $b = Module::Build->new
  (
   module_name => 'make2build',
   dist_version_from => 'make2build.pl',
   requires => {},
   recommends => {},
   build_requires => {},
   conflicts => {},
   license => 'unknown',
   create_makefile_pl => 'passthrough',
   dist_author => 'Steven Schubiger <sts@accognoscere.org>',
   add_to_cleanup => [],
  );
  
$b->create_build_script;
